--[[ 
Combat Scanner
Handles:
- Check for boss engage
- Combat log events
- Yells, emotes etc.
]]--

------------------------------------
-- Create plugin

local RW = LibStub("AceAddon-3.0"):GetAddon("Raid Watch")
local plugin = RW:Plugin("CombatScanner")
if not plugin then return end

RWW = plugin

local pName, pClass, party

function plugin:OnRegister()
	self.Zones = {}
	self.Running = {}
	
	party = RW:GetPlugin("Party")
	pName = UnitName("player")
	pClass = select(2, UnitClass("player"))

	self:RegisterEvent("ZONE_CHANGED_NEW_AREA")
	self:RegisterMessage("RW_NewMonitorZone")
	self:RegisterMessage("RW_BossRegistered")
end

function plugin:PluginEnable()
	self:RegisterEvent("ZONE_CHANGED_NEW_AREA")
	self:RegisterMessage("RW_NewMonitorZone")
	self:RegisterMessage("RW_BossRegistered")
	self:RegisterMessage("RW_TestEvent")
	self:RegisterMessage("RW_BossEngaged")
	self:RegisterMessage("RW_BossEnded")
	
	-- Call functions once in case logging in in an 
	-- enabled zone or if in combat allready when logging in
	self:ZONE_CHANGED_NEW_AREA()
	self:PLAYER_REGEN_DISABLED()
end

function plugin:PluginDisable()
	twipe(self.Zones)
	twipe(self.Running)
end

-- Scanning function to run while in combat to detect if a boss gets engaged
-- Returns the bossmod and the cID of the boss/creature that initated bossfight
local function scanEngage(self)
	if not self.Zones[GetRealZoneText()] then return end
	for bossName, boss in pairs(self.Zones[GetRealZoneText()]) do
		if boss.mobTriggers then
			for _, player in pairs(party.group) do
				for cID, _ in pairs(boss.mobTriggers) do
					if cID == boss:CUIDFromUID(player.id.."target") and UnitAffectingCombat(player.id.."target") then
						return boss, cID
					end
				end
			end
		end
	end
end

-- Scanning function while in combat with a boss to check if its a wipe
-- Triggers when all in the raid is dead or when released to ghost.
-- Wont trigger if the WinTrigger of the boss is not combat based, ie. a yell or so
local function scanWipe(self)
	local wipe = true
	for name, player in pairs(party.group) do
		if UnitAffectingCombat(player.id) and not UnitIsDeadOrGhost(player.id) then
			wipe = false
			break
		end
	end
	if not wipe then
		self:ScheduleTimer(scanWipe, 3, self)
	else
		for boss in pairs(self.Running) do
			if GetTime() - boss.engageTime > (boss.endDelay or 0) then
				self:SendMessage("RW_BossEnded", boss, boss.winTrigger ~= "combat" and true)
			else
				self:ScheduleTimer(scanWipe, 3, self)
			end
		end
	end
end
 
-- Check if in combat with any boss, else schedule to check again 
function plugin:PLAYER_REGEN_DISABLED()
	local boss, trigger = scanEngage(self)
	if boss and not self.Running[boss] then
		self:ScheduleTimer(scanWipe, 3, self)
		self:SendMessage("RW_BossEngaged", boss, trigger)
	else
		if UnitAffectingCombat("player") then 
			self:ScheduleTimer("PLAYER_REGEN_DISABLED", 2)
		end
	end
end

function plugin:MonsterCom(event, msg, name)	
	-- Trigger bossmods that are started by yells
	if self.Zones[GetRealZoneText()] then
		for bossName, boss in pairs(self.Zones[GetRealZoneText()]) do
			if boss.yellTriggers then
				for yell in pairs(boss.yellTriggers) do
					if yell == msg then
						self:ScheduleTimer(scanWipe, 3, self)
						self:SendMessage("RW_BossEngaged", boss, msg)
					end
				end
			end
		end
	end
	
	-- For all running bossmods
	for boss in pairs(self.Running) do	
		-- Trigger wins
		if boss.yellWins then
			for yell in pairs(boss.yellWins) do
				if yell == msg then
					self:SendMessage("RW_BossEnded", boss, false, msg)
				end
			end
		end		
		-- Trigger tracked yell functions
		if boss.trackedYells and boss.trackedYells[msg] then
			local func = boss.trackedYells[msg]
			if type(func) == "string" then
				boss[func](boss, event, msg, name)
			else
				func(boss, event, msg, name)
			end
		elseif boss.trackedYells then
			for yell, func in pairs(boss.trackedYells) do
				if msg:find(yell) then
					boss[func](boss, event, msg, name)
				end
			end
		end		
		-- Trigger general yell function
		if boss[event] and type(boss[event]) == "function" then
			boss[event](boss, event, msg, name)
		end
	end
end


function plugin:UNIT_DIED(gUID)
	local cID = gUID
	if type(gUID) == "string" then
		local cID = tonumber(gUID:sub(9, 12), 16)
	end
	for boss in pairs(self.Running) do
		if boss.mobWins then
			for winMob, state in pairs(boss.mobWins) do					
				if cID == winMob then
					boss.mobWins[winMob] = false
				end
				
				local win = true				
				for mob, state in pairs(boss.mobWins) do
					if state then
						win = false
						break
					end
				end				
				
				if win then
					self:SendMessage("RW_BossEnded", boss, false, "Mobs")
				end
			end
		end
		
		if boss.trackedDeaths and boss.trackedDeaths[cID] then
			local func = boss.trackedDeaths[cID]
			if type(func) == "string" then
				boss[func](boss, cID)
			else
				func(boss, cID)
			end
		end
		
		if boss["UNIT_DIED"] and type(boss["UNIT_DIED"]) == "function" then
			boss["UNIT_DIED"](boss, cID)
		end
	end
end

-- Unregister all kinds of monotoring when not in a zone with bosses that are tracked
function plugin:ZONE_CHANGED_NEW_AREA()
	local zone = GetRealZoneText()
	if self.Zones[zone] then
		self:RegisterEvent("CHAT_MSG_MONSTER_YELL", "MonsterCom")
		self:RegisterEvent("CHAT_MSG_MONSTER_EMOTE", "MonsterCom")
		self:RegisterEvent("CHAT_MSG_RAID_BOSS_EMOTE", "MonsterCom")
		self:RegisterEvent("CHAT_MSG_MONSTER_SAY", "MonsterCom")
		self:RegisterEvent("PLAYER_REGEN_DISABLED")
	else
		self:UnregisterEvent("CHAT_MSG_MONSTER_YELL")
		self:UnregisterEvent("PLAYER_REGEN_DISABLED")
		self:UnregisterEvent("CHAT_MSG_MONSTER_EMOTE")
		self:UnregisterEvent("CHAT_MSG_RAID_BOSS_EMOTE")
		self:UnregisterEvent("CHAT_MSG_MONSTER_SAY")
		self:UnregisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
	end
end

-------------------------
-- AceMessage handling --
-------------------------
function plugin:RW_BossEngaged(message, boss, trigger)
	self.Running[boss] = true
	self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
	boss:Start(trigger)
end

function plugin:RW_BossEnded(message, boss, wipe, trigger)
	if boss.mobWins then
		for mob in pairs(boss.mobWins) do
			boss.mobWins[mob] = true
		end
	end
	self.Running[boss] = nil
	self:UnregisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
	boss:End(wipe, trigger)
end

function plugin:RW_BossRegistered(message, boss)
	for i, zone in pairs(boss.loadZone) do
		self.Zones[zone] = self.Zones[zone] or {}
		self.Zones[zone][boss:GetName()] = boss
	end
	self:ZONE_CHANGED_NEW_AREA()
end

function plugin:RW_NewMonitorZone(message, zones)
	if type(zones) == "table" then
		for i, zone in pairs(zones) do
			self.Zones[zone] = self.Zones[zone] or {}
		end
	else
		self.Zones[zones] = self.Zones[zones] or {}
	end
	self:ZONE_CHANGED_NEW_AREA()
end

------------------------
-- Combatlog handling --
------------------------
do
	local lastEvent
	local lastSpell
	local argsPrototype = {}
	local mt = {__index = argsPrototype}
	local args = setmetatable({}, mt)
	RWARGS = args
	function argsPrototype:IsPlayer() if self.dName == pName then return true end end
	
	local function prepareArgs(time, event, sGUID, sName, sFlags, dGUID, dName, dFlags, ...)
		if type(args) ~= "table" then
			plugin:Msg("Error: Args table was destroyed again, last event:" ..(lastEvent or "").." spell: " ..(lastSpell or ""))
			args = setmetatable({}, mt)			
		end
		lastEvent = event
		lastSpell = ...
		wipe(args)
		args.event = event
		args.sGUID = sGUID
		args.sName = sName
		args.sFlags = sFlags
		args.dGUID = dGUID
		args.dName = dName
		args.dFlags = dFlags
		
		if event:sub(1, 5) == "SPELL" then
			args.spellID, args.spellName, args.spellSchool = ...
			if event == "SPELL_DAMAGE" then
				args.amount, args.overkill, args.school, args.resisted, args.blocked, args.absorbed, args.critical, args.glancing, args.crushing = select(4, ...)
				args.amount = args.amount - args.overkill
			elseif event == "SPELL_MISSED" then
				args.missType, args.amountMissed = select(4, ...)
			elseif event == "SPELL_MISSED" then 
				args.missType, args.amountMissed = select(4, ...)
			elseif event == "SPELL_HEAL" or event == "SPELL_BUILDING_HEAL" then 
				args.amount, args.overhealing, args.absorbed, args.critical = select(4, ...)
				args.amount = args.amount - args.overhealing
			elseif event == "SPELL_ENERGIZE" then
				args.amount, args.powerType = select(4, ...)
			elseif event:sub(1, 14) == "SPELL_PERIODIC" then
				if event == "SPELL_PERIODIC_MISSED" then
					args.missType, args.amount = select(4, ...)
				elseif event == "SPELL_PERIODIC_DAMAGE" then
					args.amount, args.overkill, args.school, args.resisted, args.blocked, args.absorbed, args.critical, args.glancing, args.crushing = select(4, ...);
					args.amount = args.amount - args.overkill
				elseif event == "SPELL_PERIODIC_HEAL" then
					args.amount, args.overhealing, args.absorbed, args.critical = select(4, ...)
					args.amount = args.amount - args.overhealing
				elseif event == "SPELL_PERIODIC_DRAIN" then
					args.amount, args.powerType, args.extraAmount = select(4, ...)
				elseif event == "SPELL_PERIODIC_LEECH" then
					args.amount, args.powerType, args.extraAmount = select(4, ...)
				elseif event == "SPELL_PERIODIC_ENERGIZE" then
					args.amount, args.powerType = select(4, ...)
				end
			elseif event == "SPELL_CAST_FAILED" then 
				args.missType = select(4, ...)
			elseif event == "SPELL_DRAIN" then
				args.amount, args.powerType, args.extraAmount = select(4, ...)
			elseif event == "SPELL_LEECH" then
				args.amount, args.powerType, args.extraAmount = select(4, ...)
			elseif event == "SPELL_INTERRUPT" then
				args.extraSpellId, args.extraSpellName, args.extraSpellSchool = select(4, ...)
			elseif event == "SPELL_EXTRA_ATTACKS" then
				args.amount = select(4, ...)
			elseif event == "SPELL_DISPEL_FAILED" then
				args.extraSpellId, args.extraSpellName, args.extraSpellSchool = select(4, ...)
			elseif event == "SPELL_DISPEL" or event == "SPELL_STOLEN" then
				args.extraSpellId, args.extraSpellName, args.extraSpellSchool, args.auraType = select(4, ...)
			elseif event == "SPELL_AURA_BROKEN" then
				args.auraType = select(4, ...)
			elseif event == "SPELL_AURA_BROKEN_SPELL" then
				args.extraSpellId, args.extraSpellName, args.extraSpellSchool, args.auraType = select(4, ...)
			elseif event == "SPELL_AURA_APPLIED" or event == "SPELL_AURA_REMOVED" or event == "SPELL_AURA_REFRESH"then
				args.auraType = select(4, ...)
			elseif ( event == "SPELL_AURA_APPLIED_DOSE" or event == "SPELL_AURA_REMOVED_DOSE" ) then
				args.auraType, args.amount = select(4, ...)
			end
		elseif event:sub(1, 5) == "SWING" then
			args.spellName = ACTION_SWING
			if event == "SWING_DAMAGE" then
				args.amount, args.overkill, args.school, args.resisted, args.blocked, args.absorbed, args.critical, args.glancing, args.crushing = ...
				args.amount = args.amount - args.overkill
			elseif event == "SWING_MISSED" then 
				args.spellName = ACTION_SWING;
				args.missType, args.amountMissed = ...
			end
		elseif event:sub(1, 5) == "RANGE" then
			args.spellId, args.spellName, args.spellSchool = ...
			if event == "RANGE_DAMAGE" then 
				args.amount, args.overkill, args.school, args.resisted, args.blocked, args.absorbed, args.critical, args.glancing, args.crushing = select(4, ...)
				args.amount = args.amount - args.overkill
			elseif event == "RANGE_MISSED" then 
				args.missType, args.amountMissed = select(4, ...)
			end
		elseif event == "DAMAGE_SHIELD" then
			args.spellId, args.spellName, args.spellSchool, args.amount, args.overkill, args.school, args.resisted, args.blocked, args.absorbed, args.critical, args.glancing, args.crushing = ...
			args.amount = args.amount - args.overkill
		elseif event == "DAMAGE_SHIELD_MISSED" then
			args.spellId, args.spellName, args.spellSchool, args.missType = ...
		elseif event == "UNIT_DIED" or event == "UNIT_DESTROYED" or event == "UNIT_DISSIPATES" then
			args.sourceName, args.sourceGUID, args.sourceFlags = args.destName, args.destGUID, args.destFlags
		elseif event == "ENVIRONMENTAL_DAMAGE" then
			environmentalType, args.amount, args.overkill, args.school, args.resisted, args.blocked, args.absorbed, args.critical, args.glancing, args.crushing = ...
			args.spellName = _G["ACTION_ENVIRONMENTAL_DAMAGE_"..environmentalType]
			args.amount = args.amount - args.overkill
		elseif event == "DAMAGE_SPLIT" then
			args.spellId, args.spellName, args.spellSchool, args.amount, args.overkill, args.school, args.resisted, args.blocked, args.absorbed, args.critical, args.glancing, args.crushing = ...;
			args.amount = args.amount - args.overkill
		end
	end

	function plugin:COMBAT_LOG_EVENT_UNFILTERED(_, time, event, sGUID, sName, sFlags, dGUID, dName, dFlags, ...)
		if event == "UNIT_DIED" then
			self:UNIT_DIED(dGUID)
		end
		local spell = select(1, ...)
		for boss in pairs(self.Running) do
			if boss.trackedSpells and boss.trackedSpells[spell] then
				prepareArgs(time, event, sGUID, sName, sFlags, dGUID, dName, dFlags, ...)
				local func = boss.trackedSpells[spell]
				if type(func) == "string" then
					boss[func](boss, event, args)
				else
					func(boss, event, args)
				end
			end
			if boss[event] and type(boss[event]) == "function" then
				prepareArgs(time, event, sGUID, sName, sFlags, dGUID, dName, dFlags, ...)
				boss[event](boss, event, args)
			end
		end
	end
	
	function plugin:COMBAT_LOG_TEST(event, args)		
		if event == "UNIT_DIED" then
			self:UNIT_DIED(args.dGUID)
		end
		for boss in pairs(self.Running) do			
			if boss.trackedSpells and boss.trackedSpells[args.spellID] then
				local func = boss.trackedSpells[args.spellID]
				if type(func) == "string" then
					boss[func](boss, event, args)
				else
					func(event, args)
				end
			end
			if boss[event] and type(boss[event]) == "function" then
				boss[event](boss, event, args)
			end
		end		
	end	
	
	function plugin:RW_TestEvent(message, event, sargs)
		args.spellID = sargs.spellID
		args.dName = sargs.dName
		args.sName = sargs.sName
		args.amount = sargs.amount
		args.spellName = sargs.spellName
		self:COMBAT_LOG_TEST(event, args)
	end
end
